// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.io.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.Debug;



/** reacts on mouse events to highlight the selected part of the
 * sequent and it pops up a menu showing all applicable Taclet 
 * at the highlighted position if the mouse is pressed
 *
 * Additionally it performs all necessary actions for draging some
 * highlighted sequent part into some other GUI component (e.g.
 * some Taclet rule instantiation dialog)
 */
class SequentViewListener extends MouseInputAdapter 
    implements DragGestureListener  {  

    private static final int LEFT_MOUSE_BUTTON = 16;
    private static final int CTRL_LEFT_MOUSE_BUTTON = 18;

    private KeYMediator mediator;
    private SequentView seqView;

    private TacletMenu menu;
    private boolean refreshHighlightning;
    private boolean modalDragNDropEnabled;


    /** last registered mouse position */
    PosInSequent mousePos = PosInSequent.createSequentPos();

    SequentViewListener(SequentView seqView,
			KeYMediator mediator) {
	this.mediator = mediator;
	this.seqView = seqView;
	menu = new TacletMenu();
	setRefreshHighlightning(true);
	setModalDragNDropEnabled(false);
    }

	
    public void mouseMoved(MouseEvent me) {	    
	if (me.getSource()==seqView && 
	    !menu.getPopupMenu().isVisible() && refreshHighlightning())  { 

	    seqView.paintHighlights(me.getPoint()); 
	}
    }
	    
    public void mousePressed(MouseEvent me) {
	if  (!menu.getPopupMenu().isVisible() && !modalDragNDropEnabled()
	     && me.getModifiers() != CTRL_LEFT_MOUSE_BUTTON) {
	    mousePos = seqView.getPosInSequent(me.getPoint()); 
	    if (mediator!= null && mousePos != null) {
		menu = new TacletMenu(seqView,
				      mediator.getFindTaclet(mousePos), 
				      mediator.getRewriteTaclet(mousePos),
				      mediator.getNoFindTaclet(),
				      mediator.getBuiltInRule(mousePos),
				      mousePos);
	    } else {
		menu = new TacletMenu(seqView,
				      SLListOfTacletApp.EMPTY_LIST,
				      SLListOfTacletApp.EMPTY_LIST,
				      SLListOfTacletApp.EMPTY_LIST,
				      SLListOfBuiltInRule.EMPTY_LIST,
				      mousePos);
	    }
		
	    menu.getPopupMenu().show(seqView,
				     me.getX(),
				     me.getY());	
	}
    }

    public void mouseReleased(MouseEvent me) {	
	if  (!modalDragNDropEnabled()) {
	    menu.setVisible(false);
	}
    }


    public void hideMenu(){
	menu.setVisible(false); 
    }


    public synchronized void setRefreshHighlightning(boolean doRefresh){
	refreshHighlightning = doRefresh;
    }
	
    public synchronized boolean refreshHighlightning(){
	return refreshHighlightning;
    }

	
    public synchronized void setModalDragNDropEnabled(boolean allowDragNDrop){
	modalDragNDropEnabled = allowDragNDrop;
    }
	
    public synchronized boolean modalDragNDropEnabled(){
	return modalDragNDropEnabled;
    }
	
    protected PosInSequent getMousePos() {
	return mousePos;
    }

    /**
     * a drag gesture has been initiated
     */
	
    public void dragGestureRecognized(DragGestureEvent dgEvent) {

	if (modalDragNDropEnabled() ||
	    ((MouseEvent)dgEvent.getTriggerEvent()).getModifiers() == 
	    CTRL_LEFT_MOUSE_BUTTON ){
		
	    // prevent from changing the current highlightning 
	    // (until the drag operation has been finished !)
	    setRefreshHighlightning(false);
		    
	    hideMenu();
		    
	    Point dragOrigin = dgEvent.getDragOrigin();
	    PosInSequent localMousePos = seqView.getPosInSequent(dragOrigin);
		    
	    String subtermString = null;
	    if (localMousePos != null) {
		// a subterm has been selected ...
		try {
		    subtermString = seqView.getText
			(localMousePos.getBounds().start(), 
			 localMousePos.getBounds().length());		
		} catch (BadLocationException ble){
		    ble.printStackTrace();
		}
			
	    }
		    
	    if ( subtermString != null ){
		StringSelection text = new StringSelection( subtermString ); 
		
		// as the name suggests, starts the dragging
		seqView.getDragSource().startDrag (dgEvent, 
				      DragSource.DefaultCopyDrop, 
				      text, 
				      new ADragSourceListener());
	    } 
	}
    }
	

    class ADragSourceListener implements  DragSourceListener {
						  
	/**
	 * is invoked when you are dragging over the DropSite
	 * 
	 */
						  
	public void dragEnter (DragSourceDragEvent event) {
	}
						  
	/**
	 * is invoked when you are exit the DropSite without dropping
	 *
	 */
						  
	public void dragExit (DragSourceEvent event) {
	}
						  
	/**
	 * is invoked when a drag operation is going on
	 * 
	 */
						  
	public void dragOver (DragSourceDragEvent event) {
	}
						  
						  
	/**
	 * this message goes to DragSourceListener, informing 
	 * it that the dragging has ended
	 * 
	 */
						  
	public void dragDropEnd (DragSourceDropEvent event) {   
	    // Enable updating the subterm highlightning ...
	    setRefreshHighlightning(true);						
						      
	}						     
						  
	/**
	 * is invoked if the use modifies the current drop gesture
	 * 
	 */
						  
	public void dropActionChanged ( DragSourceDragEvent event ) {
	}
						  
    }

}
