// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

package de.uka.ilkd.key.parser;

import java.io.*;
import antlr.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.RuleSource;
import de.uka.ilkd.key.proof.init.Includes;
import de.uka.ilkd.key.gui.AbbrevMap;

import de.uka.ilkd.key.rule.metaconstruct.*;
import de.uka.ilkd.key.java.JavaReader;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;

import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import java.util.List;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.HashMap;
import de.uka.ilkd.key.java.visitor.DeclarationProgramVariableCollector;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
}

/** 
 * A parser for first-order terms and formulae.
 */  
class TermParser extends DeclParser;
options {
    importVocab=KeYLexer;
    k = 1;
    defaultErrorHandler=false;
}


{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected Choice defaultChoice = null;
    protected HashMap category2Default = new HashMap();
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();
    protected NamespaceSet nss;

    private Prioritization.StaticPrioritization heur_prioritization
        = new Prioritization.StaticPrioritization ();


    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    private boolean skip_schemavariables;
    private boolean skip_functions;
    private boolean skip_predicates;
    private boolean skip_sorts;
    private boolean skip_heuristics;

    private Services services;
    private AbbrevMap scm;

    private int lineOffset=0;
    private int colOffset=0;

    private TermFactory tf;
    private JavaReader javaReader;
    
    private boolean errors = false;

    public TermParser(TokenStream  lexer,                   
                      String       filename,
                      TermFactory  tf,
                      JavaReader   jr,
                      Services     services,
                      NamespaceSet nss,
                      AbbrevMap    scm) {
        this(lexer,2);
        setFilename(filename);
        this.javaReader = jr;
        this.nss = nss;
        this.tf  = tf;
        this.services = services;
        this.scm = scm;
        switchToNormalMode();
    }

    protected int getLine() {
        int line = -1;
        try {
            line = LT(0).getLine()+lineOffset;
        } catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return line;
    }   

    public JavaInfo getJavaInfo() {
        return getServices().getJavaInfo();
    }

    public Services getServices() {
        return services;
    }

    private Namespace sorts() {
        return nss.sorts();
    }

    private Namespace variables() {
        return nss.variables();
    }

    private Namespace programVariables() {
        return nss.programVariables();
    }

    private Namespace functions() {
        return nss.functions();
    }

    private Namespace heuristics() {
        return nss.heuristics();
    }

    private Namespace choices(){
        return nss.choices();
    }

    private NamespaceSet namespaces() {
        return nss;
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};
	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    // a pure term parser is never in schemamode
    // only grammers that extend this grammar may 
    // activate the schema mode

    public void switchToSchemaMode() {
    }

    private boolean inSchemaMode() {
        return false;
    }

    public void switchToNormalMode() {
    }


    public Term toZNotation(String s){
        Term result = tf.createFunctionTerm
	    ( (Function) functions().lookup(new Name(s.substring(0,1))),
	      tf.createFunctionTerm((Function) functions().lookup(new Name("#"))));

        for(int i=1;i<s.length();i++){
	    result = tf.createFunctionTerm
		((Function)functions().lookup
		  (new Name(s.substring(i,i+1))), result);
        }
        return tf.createFunctionTerm
	    ((Function) functions().lookup(new Name("Z")), result);	
    }


    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                                   filename+"\n " +e);
            }
            filename = filename.substring(
                                          filename.lastIndexOf(File.separator)+1,
                                          filename.lastIndexOf(".key"));
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }  

    protected int getColumn() {
        int     col = -1;
        try {
            col = LT(0).getColumn()+colOffset;
        } 
        catch (TokenStreamException e) {
            System.err.println("No further token in stream");
        }
        return col;
    }   

    protected LogicVariable bindVar(Token id,Sort s) {
        LogicVariable v=new LogicVariable(new Name(id.getText()), s);
        namespaces().setVariables(variables().extended(v));
        return v;
    }

    protected TermSymbol progVar(Token id, Sort s) throws NotDeclException{
        TermSymbol v = (TermSymbol)
            variables().lookup(new Name(id.getText()));

        if (v == null) {
            throw new NotDeclException("program variable", id,
                                       getFilename());
        }
        return v;
    }

    protected boolean isNumber(String s){
        boolean num=true;
        if(s==null){
            num=false;
        }else{
            for(int i=s.length()-1;i>=0;i--){
                if(s.charAt(i)<'0'||s.charAt(i)>'9')
                    num=false;
            }
        }
        return num;
    }

    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }
    
    protected void unbindVars() {
        namespaces().setVariables(variables().parent());
    }

    protected void bindProgVars(HashSet progVars) {
        namespaces().setVariables(new Namespace(variables()));
        Iterator it=progVars.iterator();
        while (it.hasNext()) {
            variables().add((IProgramVariable)it.next());
        }
    }

    protected void unbindProgVars() {
        namespaces().setVariables(variables().parent());
    }

    protected static HashSet progVars(JavaBlock jb) {
        if (jb==JavaBlock.EMPTY_JAVABLOCK) {
            return new HashSet();
        }   
        DeclarationProgramVariableCollector pvc
            = new DeclarationProgramVariableCollector(jb.program());
        pvc.start();
        return pvc.result();
    }


    protected Term termForParsedVariable(ParsableVariable v) 
        throws antlr.SemanticException {
        if ( v instanceof LogicVariable ) {
            return tf.createVariableTerm((LogicVariable)v);
        } else if ( v instanceof ProgramVariable ) {
            return tf.createVariableTerm((ProgramVariable)v);
        } else if ( v instanceof Metavariable ) {
            return tf.createFunctionTerm((Metavariable)v);
        } else {
            throw new antlr.SemanticException(
                                              ""+ v +" is not a logic variable");
        }
    }
    
    protected JavaBlock getJavaBlock(Token t) throws antlr.SemanticException {
        String s=t.getText();
        try {
            return javaReader.readBlockWithProgramVariables(variables(), s);
        } catch (de.uka.ilkd.key.java.PosConvertException e) {
            lineOffset=e.getLine()-1;
            colOffset=e.getColumn()+1;
            throw new JavaParserException(e.getMessage(), t, 
                                          getFilename(), lineOffset, colOffset);                        
        } catch (de.uka.ilkd.key.java.ConvertException e) { 
            if (e.parseException()!=null) {               
                lineOffset=e.parseException().currentToken.next.beginLine;               
                colOffset=e.parseException().currentToken.next.beginColumn;
                e.parseException().currentToken.next.beginLine=getLine()-1;
                e.parseException().currentToken.next.beginColumn=getColumn();
                throw new JavaParserException(e.parseException().getMessage(), t, 
                                              getFilename(), -1, -1);  // row/columns already in text
            }       
            throw new JavaParserException(e.getMessage(), t, getFilename(), 0, 1);
        }
    }

}

/**
 * In the special but important case of Taclets, we don't yet know
 * whether we are going to have a term or a formula, and it is hard
 * to decide a priori what we are looking at.  So the `term' 
 * non-terminal will recognize a term or a formula.  The `formula'
 * reads a formula/term and throws an error if it wasn't a formula.
 * This gives a rather late error message. */

top {Term t;} : t=formula;

formula returns [Term a = null] 
    :
        a = term 
        {
            if ( a.sort() != Sort.FORMULA ) {
                throw 
                new antlr.SemanticException
                ("Just Parsed a Term where a Formula was expected");
            }
        }
    ;


term returns [Term a = null] 
    :
        a=term20 
    ;

term20 returns [Term a = null] 
{
    Term a1;
}
    :   a=term30 
        (EQV a1=term20 
            { a = tf.createJunctorTerm(Op.EQV, new Term[]{a, a1});} ) ?;
term30 returns [Term a = null] 
{
    Term a1;
}
    :   a=term40 
        (IMP a1=term30 
            { a = tf.createJunctorTerm(Op.IMP, new Term[]{a, a1});} ) ?;
term40 returns [Term a = null] 
{
    Term a1;
}
    :   a=term50 
        (OR a1=term40 
            { a = tf.createJunctorTerm(Op.OR, new Term[]{a, a1});} ) ?;
term50 returns [Term a = null] 
{
    Term a1;
}
    :   a=term60 
        (AND a1=term50
            { a = tf.createJunctorTerm(Op.AND, new Term[]{a, a1});} ) ?
    ;



term60 returns [Term a = null] 
{
    Term a1,a2;
    Sort s;
}
    :  
        NOT a1  = term60 { a = tf.createJunctorTerm(Op.NOT,new Term[]{a1}); }
    |   CTRL a1 = term60 { a = tf.createJunctorTerm(Op.CTRL,new Term[]{a1}); }
    |   a = bindingterm 
    |   a = dlTermOrTerm70
    ;

/**
 * this is a hook that can be redefined by inheriting grammars to
 * redefine the sorts of things that can appear inside a diamond or
 * box.  In particular, in Taclets, schema variables and pre/suffixes
 * might occur. 
 * And to get the parser deterministic it passes through a term70
 */
dlTermOrTerm70 returns [Term a = null] 
{}
    :
        a = normal_dl_term
    |  (IDENT MODALITYARG)=> a = named_dl_term
    |  a = term70
    ; 


term70 returns [Term a = null] 
{
    Term a1;
}
    :
        a = accessterm
        // a term like {o:=u}x=y is parsed as {o:=u}(x=y)
        ((EQUALS)=>EQUALS a1 = logicTermReEntry
            { a = tf.createEqualityTerm(a, a1);}) ?

    ;

/**
 * helps to better distinguish if formulas are allowed or only term are
 * accepted
 */
logicTermReEntry returns [Term result = null]
{}
    :
        (
            result = accessterm |
            result = update_or_substitution
        ) 
        {
            if (result.sort() == Sort.FORMULA) {
                throw 
                    new KeYSemanticException
                    ("Only terms may stand here.", 
                     getLine(), getColumn(), 
                     getFilename());
            }
        }          
    ;

accessterm returns [Term result = null] 
{
    KeYJavaType kjt = null;
    Sort arraySort = null;
    Term indexTerm = null;
    int shadowPrimes = 0;
}
    :
        (LPAREN logicTermReEntry (COMMA | GUARD)) => result = conditional
    |
        {"-".equals(LT(1).getText())}? neg:ARITH_OP result = logicTermReEntry
        {
            if (result.sort() != Sort.FORMULA) {
                result = tf.createFunctionTerm
                ((Function) functions().lookup(new Name("~m")), result);
            } else {
                throw new KeYSemanticException
                ("Formula cannot be prefixed with: ", neg, getFilename());
            }
        } 
    | {LA(2) == DOT && ((kjt = getTypeByClassName(LT(1).getText()))!=null)}?
        name:IDENT {            
            result = tf.createFunctionTerm 
                (new ProgramVariable 
                 (new ProgramElementName(name.getText()), kjt));        
        }
        result = attribute_suffix[result]
    | 
        result = term80
        (LBRACKET indexTerm = logicTermReEntry
            ( RBRACKET
            | rb:RBRACKETPRIMES {
                    shadowPrimes = rb.getText().length()-1;                 
                }
            )
            {
                if (arraySort != null && 
                    arraySort != ProgramSVSort.VARIABLE) {
                        if (!(arraySort instanceof ArraySort)){
                            throw new NotDeclException
                            ("array "+result+" declared with a different"+
                             " dimension than required here", ""+result, 
                             getFilename(), getLine(), getColumn());
                        } else {
                            arraySort = ((ArraySort) arraySort).elementSort();
                        } 
                    } else if (arraySort == ProgramSVSort.VARIABLE && !inSchemaMode()) {
                        throw new antlr.SemanticException
                        ("Schemavariable are not allowed in this context.");
                    }
                if (shadowPrimes>0) {
                    result = tf.createArrayTerm
                    (Op.SHADOW_ARRAY, result, indexTerm, 
                     toZNotation("" + shadowPrimes));
                } else {
                    result = tf.createArrayTerm(Op.ARRAY, result, indexTerm);
                }   
            }
        | result = attribute_suffix[result]
        )*
    ;


attribute_suffix[Term prefix] returns [Term result = null]
{
    TermSymbol v = null;
    int shadowPrimes = 0;
    result = prefix;
}    
    :   
        DOT v = attrid[result]
        (ps:PRIMES 
            { shadowPrimes = ps.getText().length();})?
        {       
            Sort so = v.sort(); 
            if (!inSchemaMode()) {
                if (((ProgramVariable)v).isStatic()){
                    result = tf.createVariableTerm((ProgramVariable)v);
                } else {
                    if (shadowPrimes > 0) {
                        Term shadownum = toZNotation("" + shadowPrimes);
                        result = tf.createShadowAttributeTerm((ProgramVariable)v, 
                                                              result, shadownum);
                    } else {
                        result = tf.createAttributeTerm((ProgramVariable)v, result);
                    }
                }
            } else {
                if(shadowPrimes>0){
                    Term shadownum = toZNotation("" + shadowPrimes);
                    result = tf.createShadowAttributeTerm((SchemaVariable)v, result, shadownum);         
                } else
                result = tf.createAttributeTerm((SchemaVariable)v, result);         
            }
        }
    ;

term80 returns [Term a = null]
    :
        a = funcpredvarterm
    |   a = specialTerm
    |   LPAREN a = term RPAREN 
    |   "true"  { a = tf.createJunctorTerm(Op.TRUE); }
    |   "false" { a = tf.createJunctorTerm(Op.FALSE); }
    ;
//    exception // for rule
//      catch [RecognitionException ex] {
//        throw new RecognitionException("Not a valid logical term.\n"+ex.toString());
//    }



abbreviation returns [Term a=null]
{ 
    String sc = null;
}
    :
        (   sc = id
            {
                a =  scm.getTerm(sc);
                if(a==null){
                    throw new NotDeclException
                        ("abbreviation", sc, 
                         getFilename(), getLine(), getColumn());
                }                                
            }
        )
    ;

conditional returns [Term result = null] 
{ 
    Term comparator = null;
}
    :
        LPAREN comparator = logicTermReEntry 
        ( result=normalConditional[comparator] | result=conjunctionConditional[comparator])
    ;

normalConditional[Term cond1] returns [Term result = null] 
{ 
    Term cond2 = null;
    Term c_then = null;
    Term c_else = null;
}
    :
        GUARD cond2 = logicTermReEntry RPAREN
        (
            DOT result = shortCond[cond1, cond2]
        | 
            c_then = logicTermReEntry COLON c_else = logicTermReEntry {
                result = tf.createIfElseTerm(cond1, cond2, c_then, c_else);
            }

        ) 
    ;

// shortCond is overwritten in taclets.g
shortCond[Term cond1, Term cond2] returns [Term result = null]
{
    ParsableVariable v=null;
    Term c_then = null;
}:
        DOT v = varId CONDUP c_then = logicTermReEntry {
            result = tf.createIfElseTerm((ProgramVariable)v, cond1, 
                                         cond2, c_then);
        }
    ;


conjunctionConditional[Term cond1_1] returns [Term a = null]
{ 
    Term cond1_2 = null;
    Term cond2_1 = null;
    Term cond2_2 = null;
    Term c_then = null;
    Term c_else = null;
} :
        COMMA cond1_2=logicTermReEntry COLON c_else=logicTermReEntry 
        GUARD cond2_1 = logicTermReEntry 
        COMMA cond2_2=logicTermReEntry COLON c_then=logicTermReEntry 
        RPAREN
        {
            a = tf.createConjCondTerm(new Term[]{cond1_1, cond1_2, c_else, 
                                                 cond2_1, cond2_2, c_then});
        }        
    ;


/**
 * this is a hook that can be redefined by inheriting grammars to
 * introduce other sorts of binding behaviour for quantifiers or
 * substitution terms.  In particular, the Taclet parser will hook in
 * here to allow quantification of schema variables.
 */
bindingterm returns [Term a = null]
    :
        a = local_bindingterm             
    ;


/** here is the standard behaviour, that declares a new variable
  * with local scope in a quantifier */

local_bindingterm returns [Term a = null]
{
    Operator op = null;
    LogicVariable v = null;
    Term a1;
    Sort s;
}
    :
        (
            "all" { op = Op.ALL; }
        |   "ex"  { op = Op.EX;  }
        )
        v0:IDENT COLON s=sortId DOT {v = bindVar(v0,s);} a1 = term60
        {
            a = tf.createQuantifierTerm((Quantifier)op, v,a1);
            unbindVars();
        }
    | a = update_or_substitution
    ;


update_or_substitution returns [Term result = null]
{
} :

        ( LBRACE IDENT COLON ) => result = substitutionterm |
        result = updateterm
    ; 


substitutionterm returns [Term result = null]
{ Sort s;
LogicVariable v = null;
Term a1, a2;
boolean waryAll = false;
boolean waryEx = false;
} : 

        LBRACE v1:IDENT COLON s=sortId a1=logicTermReEntry
        ( "waryAll" { waryAll = true; } |
            "waryEx"  { waryEx  = true; } )?
        RBRACE 
        {v = bindVar(v1, s);} 
        a2 = term60 { 
            if ( waryAll )
            result = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
                                                 v, a1, a2 );
            else if ( waryEx )
            result = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
                                                 v, a1, a2 );
            else
            result = tf.createSubstitutionTerm ( Op.SUBST,
                                                 v, a1, a2 );
            unbindVars();
        }
    ;



updateterm returns [Term result = null] 
{ 
    SingleUpdateData sud = null;
    List subs = new LinkedList();  
    Term a2 = null;
} :
        LBRACE
        sud = singleupdate {
            subs.add(sud.a0); 
            subs.add(sud.a1); 
        } ( COMMA sud = singleupdate {
                subs.add(sud.a0); 
                subs.add(sud.a1); 
            } )* 
        RBRACE a2 = term60 {
            subs.add(a2);
        }
        { 
            result = tf.createUpdateTerm((Term[])subs.toArray(new Term[subs.size()]));
        }
    ;


singleupdate returns[SingleUpdateData sud=null]
{
    Term a0 = null;
    Term a1 = null;
    ArraySort as;
    TermSymbol pv;
    LhsUpdateData lud = null;
    boolean shadow_l = false;
    boolean shadow_r = false;
}  :
        (a0 = lhsSingle ) ASSIGN (a1 = rhsSingle) {
            sud = new SingleUpdateData();
            sud.a0 = a0;
            sud.a1 = a1;
        }
    ;

rhsSingle returns[Term result = null]
{} :
        result = logicTermReEntry
    ;

lhsSingle returns[Term result = null]
{
    Term a1 = null;
} :
        result = logicTermReEntry
        { 
            if (!(result.sort() instanceof ProgramSVSort) &&
                result.op() != Op.META_SHADOW &&
                !(result.op() instanceof ProgramVariable || 
                  result.op() instanceof AccessOp)) {
                      throw new KeYSemanticException
                      ("Term of sort " + result.sort() + 
                       " cannot stand for a location.",
                       getLine(), getColumn(), getFilename());
                  }
        }
    ;



/** here is the standard behaviour, only accepts a program. */

normal_dl_term returns [Term a = null]
{   JavaBlock jb=null;
Term a1;
}
    :
        d1:DIAMOND 
        {
            jb=getJavaBlock(d1);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createDiamondTerm(jb, a1);
            unbindProgVars();
        }
    |
        d2:BOX
        {
            jb=getJavaBlock(d2);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60   
        {
            a = tf.createBoxTerm(jb, a1);
            unbindProgVars();
        }
    |
        d3:THROUGHOUT
        {
            jb=getJavaBlock(d3);
            HashSet progVars=progVars(jb);
            bindProgVars(progVars);
        }
        a1=term60
        {
            a = tf.createProgramTerm(Op.TOUT,jb, a1);
            unbindProgVars();
        }
    ;


named_dl_term returns [Term a = null]
{
    Term[] terms = null;
    Term term = null;
    JavaBlock jb = null;
} :
        id:IDENT
        bl:MODALITYARG (
            terms = argument_list
            {
                Operator op = null;
                jb = getJavaBlock(bl);
                if (id.getText().charAt(0)=='#') {
                    if (!inSchemaMode()) {
                        throw new KeYSemanticException
                            ("No schema elements allowed outside taclet declarations.",
                             id, getFilename());
                    }
                    op = (SchemaVariable)variables().lookup(new Name(id.getText()));
                } else {
                    op = Op.getModality(id.getText());
                }
                if(op == null) {
                    throw new KeYSemanticException("Unknown modal operator.", id, getFilename());
                }
                if(op.arity() != terms.length) {
                    throw new KeYSemanticException
                    ("The arity of " + op.name() +
                     " is "+op.arity() + ", but you provided " + terms.length + " arguments",
                     getLine(), getColumn(), getFilename());
                }
                a = tf.createProgramTerm(op, jb, terms);
            }
        )
    ;


argument_list returns [Term[] ts = null]
{
    List args = new LinkedList();
    Term a1, a2;
}
    :
        LPAREN 
        a1 = term { args.add(a1); }

        (COMMA a2 = term { args.add(a2); } )* 

        RPAREN
        {
            ts = (Term[]) args.toArray(new Term[args.size()]);
        }
    ;

funcpredvarterm returns [Term a = null]
{
    Operator vf=null;
    Term[] arguments = AN_ARRAY_OF_TERMS;
}
    :
        ch:CHAR_LITERAL {
            String s = ch.getText();
            int intVal = 0;
            if (s.length()==3) {
                intVal = (int)s.charAt(1);
            } else {
                try {
                    intVal = Integer.parseInt(s.substring(3,s.length()-1),16);
                } catch (NumberFormatException ex) {
                    throw new KeYSemanticException
                        ("Expected '"+s+"' to be a character", ch, getFilename());
                }       
            }
            a = tf.createFunctionTerm((Function) functions().lookup(new Name("C")), 
                                      toZNotation(""+intVal).sub(0));
        }
    | {isNumber(LT(1).getText()) && !"(".equals(LT(2).getText())}?       
        number:IDENT
        { a = toZNotation(number.getText());}    
    | AT a = abbreviation
    | vf = varfuncId ((LPAREN)=>arguments = argument_list)?
        {   
            if (vf instanceof ParsableVariable) {
                a = termForParsedVariable((ParsableVariable)vf);
            } else {
                a = tf.createFunctionTerm((Function)vf, arguments);
            } 
        }
    ;


attrid [Term prefix] returns [TermSymbol v = null]
{} : 
        attr:IDENT     
        {
            v = getJavaInfo().getAttribute(attr.getText(), (ObjectSort)prefix.sort());
            if ( v == null && !("length").equals(attr.getText()) ) {
                throw new NotDeclException
                    ("Attribute ", attr.getText(), getFilename(), 
                     getLine(), getColumn());

            }  
        };

varfuncId returns [Operator v = null] 
{ 
    String varfunc_name = null;
}
    :
        varfunc_name = id       
        {                        
            v = (TermSymbol) variables().lookup(new Name(varfunc_name));
            if (v == null) {
                v = (Function) functions().lookup(new Name(varfunc_name));
                if (v == null) {
                    v = (ProgramVariable)namespaces().
                        programVariables().lookup(new Name(varfunc_name));
                    if (v == null) {
                        throw new NotDeclException
                            ("variable or function", varfunc_name, 
                             getFilename(), getLine(), getColumn());
                    }                 
                }
            }
        }
    ;


// entry point overwritten in taclets.g
specialTerm returns [Term result = null] 
{
    Operator vf = null;
}:
        vf = expr_op 
        {   
            result = tf.createFunctionTerm((Function)vf, AN_ARRAY_OF_TERMS);
        }                 
    ;

expr_op returns [Operator v = null] 
{
    ParsableVariable left_var0 = null;
    ParsableVariable right_var0 = null;
    ParsableVariable unary_var0 = null;
}   
    :

        IN_TYPE LPAREN ((left_var0=varId ((id:ARITH_OP | id1:SLASH | id2:IDENT) right_var0 = varId)?) |
            ( "-" unary_var0 = varId) )  RPAREN
        {  
            if (id==null) id=id1;   
            if (unary_var0!=null) {
                // can only be unary minus
                SortedSchemaVariable unary_var = (SortedSchemaVariable) unary_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                return new InTypeOperator
                    (type, new de.uka.ilkd.key.java.expression.operator.Negative(unary_var));
            }
            else {
                SortedSchemaVariable left_var = (SortedSchemaVariable) left_var0;
                Sort type = de.uka.ilkd.key.logic.sort.Sort.FORMULA;
                if (right_var0 != null) {
                    SortedSchemaVariable right_var = (SortedSchemaVariable) right_var0;
                    String op_str = (id == null ? id2.getText() : id.getText());
                    if ("+".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Plus(left_var, right_var));
                    } else if ("-".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Minus(left_var, right_var));
                    } else if ("*".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Times(left_var, right_var));
                    } else if ("/".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Divide(left_var, right_var));
                    } else if ("%".equals(op_str)) {
                        return new InTypeOperator
                            (type, new de.uka.ilkd.key.java.expression.operator.Modulo(left_var, right_var));
                    } 
                }
                else {
                    return new InTypeOperator(type, left_var);
                }
            }
        }
    | IS_OBJECT  LPAREN (left_var0=varId )  RPAREN
        {
            SortedSchemaVariable var = (SortedSchemaVariable) left_var0;
            return new IsObject(de.uka.ilkd.key.logic.sort.Sort.FORMULA, var);
        }
    ;

varId returns [ParsableVariable v = null]
    :
        id:IDENT 
        {   
            v = (ParsableVariable) variables().lookup(new Name(id.getText()));
            if (v == null) {
                throw new NotDeclException("variable", id, 
                                           getFilename());
            }
        }
    ;
